NAMESPACES_INTERNAL = ["Internal", "Profile", "Template"]

schema_attribute_type_mapping = {
    "Text": "String",
    "Number": "Integer",
    "Boolean": "Boolean",
    "DateTime": "DateTime",
    "Enum": "String",
}



import json
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any

from agents import Agent
from agents.result import RunResult
from deepeval.test_case.llm_test_case import ToolCall
from openai.types.responses import (
    ResponseFunctionToolCall,
)

CURRENT_DIRECTORY = Path(__file__).parent.resolve()
ROOT_DIRECTORY = CURRENT_DIRECTORY.parent.resolve()


@asynccontextmanager
async def agent_context(**kwargs: Any) -> AsyncIterator[Agent]:
    """Context manager to create an Agent with MCP server.

    The Context manager will connect the MCP server(s) and cleanup after the agent is done.
    """
    agent = Agent(**kwargs)
    # Connect all MCP servers
    for mcp_server in agent.mcp_servers:
        await mcp_server.connect()
    try:
        yield agent
    finally:
        # Cleanup all MCP servers
        for mcp_server in agent.mcp_servers:
            await mcp_server.cleanup()


def extract_tools(result: RunResult) -> list[ToolCall]:
    """Extract tool calls during an Agent run from the result."""
    tools: list[ToolCall] = []
    for response in result.raw_responses:
        for item in response.output:
            if isinstance(item, ResponseFunctionToolCall) and item.status == "completed":
                tool_call = ToolCall(
                    name=item.name, input_parameters=json.loads(item.arguments) if item.arguments else None
                )
                tools.append(tool_call)
    return tools



from typing import TYPE_CHECKING, Annotated, Any

from fastmcp import Context, FastMCP
from mcp.types import ToolAnnotations
from pydantic import Field

from infrahub_mcp.utils import MCPResponse, MCPToolStatus

if TYPE_CHECKING:
    from infrahub_sdk import InfrahubClient

mcp: FastMCP = FastMCP(name="Infrahub GraphQL")


@mcp.tool(tags={"schemas", "retrieve"}, annotations=ToolAnnotations(readOnlyHint=True))
async def get_graphql_schema(ctx: Context) -> MCPResponse:
    """Retrieve the GraphQL schema from Infrahub

    Parameters:
        None

    Returns:
        MCPResponse with the GraphQL schema as a string.
    """
    client: InfrahubClient = ctx.request_context.lifespan_context.client
    resp = await client._get(url=f"{client.address}/schema.graphql")  # noqa: SLF001
    return MCPResponse(status=MCPToolStatus.SUCCESS, data=resp.text)


@mcp.tool(tags={"schemas", "retrieve"}, annotations=ToolAnnotations(readOnlyHint=False))
async def query_graphql(
    ctx: Context, query: Annotated[str, Field(description="GraphQL query to execute.")]
) -> MCPResponse[dict[str, Any]]:
    """Execute a GraphQL query against Infrahub.

    Parameters:
        query: GraphQL query to execute.

    Returns:
        MCPResponse with the result of the query.

    """
    client: InfrahubClient = ctx.request_context.lifespan_context.client
    data = await client.execute_graphql(query=query)

    return MCPResponse(status=MCPToolStatus.SUCCESS, data=data)



from typing import TYPE_CHECKING, Annotated

from fastmcp import Context, FastMCP
from infrahub_sdk.branch import BranchData
from infrahub_sdk.exceptions import GraphQLError
from mcp.types import ToolAnnotations
from pydantic import Field

from infrahub_mcp.utils import MCPResponse, MCPToolStatus, _log_and_return_error

if TYPE_CHECKING:
    from infrahub_sdk import InfrahubClient

mcp: FastMCP = FastMCP(name="Infrahub Branches")


@mcp.tool(
    tags={"branches", "create"},
    annotations=ToolAnnotations(readOnlyHint=False, idempotentHint=True, destructiveHint=False),
)
async def branch_create(
    ctx: Context,
    name: Annotated[str, Field(description="Name of the branch to create.")],
    sync_with_git: Annotated[bool, Field(default=False, description="Whether to sync the branch with git.")],
) -> MCPResponse:
    """Create a new branch in infrahub.

    Parameters:
        name: Name of the branch to create.
        sync_with_git: Whether to sync the branch with git. Defaults to False.

    Returns:
        Dictionary with success status and branch details.
    """

    client: InfrahubClient = ctx.request_context.lifespan_context.client
    await ctx.info(f"Creating branch {name} in Infrahub...")

    try:
        branch = await client.branch.create(branch_name=name, sync_with_git=sync_with_git, background_execution=False)

    except GraphQLError as exc:
        return await _log_and_return_error(ctx=ctx, error=exc, remediation="Check the branch name or your permissions.")

    return MCPResponse(
        status=MCPToolStatus.SUCCESS,
        data={
            "name": branch.name,
            "id": branch.id,
        },
    )


@mcp.tool(tags={"branches", "retrieve"}, annotations=ToolAnnotations(readOnlyHint=True))
async def get_branches(ctx: Context) -> MCPResponse:
    """Retrieve all branches from infrahub."""

    client: InfrahubClient = ctx.request_context.lifespan_context.client
    await ctx.info("Fetching all branches from Infrahub...")

    branches: dict[str, BranchData] = await client.branch.all()

    return MCPResponse(status=MCPToolStatus.SUCCESS, data=branches)



import pytest


@pytest.fixture
def locationsite_filters() -> dict[str, str]:
    return {
        "contact__value": "String",
        "address__value": "String",
        "city__value": "String",
        "name__value": "String",
        "description__value": "String",
        "devices__type__value": "String",
        "devices__name__value": "String",
        "devices__role__value": "String",
        "devices__description__value": "String",
        "devices__status__value": "String",
        "subscriber_of_groups__name__value": "String",
        "subscriber_of_groups__group_type__value": "String",
        "subscriber_of_groups__description__value": "String",
        "subscriber_of_groups__label__value": "String",
        "circuit_endpoints__description__value": "String",
        "profiles__profile_name__value": "String",
        "profiles__profile_priority__value": "Integer",
        "member_of_groups__name__value": "String",
        "member_of_groups__group_type__value": "String",
        "member_of_groups__description__value": "String",
        "member_of_groups__label__value": "String",
        "tags__description__value": "String",
        "tags__name__value": "String",
        "parent__name__value": "String",
        "parent__description__value": "String",
        "children__role__value": "String",
        "children__facility_id__value": "String",
        "children__serial_number__value": "String",
        "children__height__value": "String",
        "children__status__value": "String",
        "children__asset_tag__value": "String",
        "children__name__value": "String",
        "children__description__value": "String",
        "vlans__name__value": "String",
        "vlans__status__value": "String",
        "vlans__role__value": "String",
        "vlans__vlan_id__value": "Integer",
        "vlans__description__value": "String",
    }



from typing import TYPE_CHECKING, Annotated, Any

from fastmcp import Context, FastMCP
from infrahub_sdk.exceptions import BranchNotFoundError, SchemaNotFoundError
from mcp.types import ToolAnnotations
from pydantic import Field

from infrahub_mcp.constants import NAMESPACES_INTERNAL
from infrahub_mcp.utils import MCPResponse, MCPToolStatus, _log_and_return_error

if TYPE_CHECKING:
    from infrahub_sdk import InfrahubClient

mcp: FastMCP = FastMCP(name="Infrahub Schemas")


@mcp.tool(tags={"schemas", "retrieve"}, annotations=ToolAnnotations(readOnlyHint=True))
async def get_schema_mapping(
    ctx: Context,
    branch: Annotated[
        str | None,
        Field(default=None, description="Branch to retrieve the mapping from. Defaults to None (uses default branch)."),
    ],
) -> MCPResponse:
    """List all schema nodes and generics available in Infrahub

    Parameters:
        branch: Branch to retrieve the mapping from. Defaults to None (uses default branch).

    Returns:
        Dictionary with success status and schema mapping.
    """
    client: InfrahubClient = ctx.request_context.lifespan_context.client
    if branch:
        await ctx.info(f"Fetching schema mapping for {branch} from Infrahub...")
    else:
        await ctx.info("Fetching schema mapping from Infrahub...")

    try:
        all_schemas = await client.schema.all(branch=branch)
    except BranchNotFoundError as exc:
        return await _log_and_return_error(ctx=ctx, error=exc, remediation="Check the branch name or your permissions.")

    # TODO: Should we add the description ?
    schema_mapping = {
        kind: node.label or "" for kind, node in all_schemas.items() if node.namespace not in NAMESPACES_INTERNAL
    }

    return MCPResponse(
        status=MCPToolStatus.SUCCESS,
        data=schema_mapping,
    )


@mcp.tool(tags={"schemas", "retrieve"}, annotations=ToolAnnotations(readOnlyHint=True))
async def get_schema(
    ctx: Context,
    kind: Annotated[str, Field(description="Schema Kind to retrieve.")],
    branch: Annotated[
        str | None,
        Field(default=None, description="Branch to retrieve the schema from. Defaults to None (uses default branch)."),
    ],
) -> MCPResponse:
    """Retrieve the full schema for a specific kind.
    This includes attributes, relationships, and their types.

    Parameters:
        kind: Schema Kind to retrieve.
        branch: Branch to retrieve the schema from. Defaults to None (uses default branch).

    Returns:
        Dictionary with success status and schema.
    """
    client: InfrahubClient = ctx.request_context.lifespan_context.client
    await ctx.info(f"Fetching schema of {kind} from Infrahub...")

    try:
        schema = await client.schema.get(kind=kind, branch=branch)
    except SchemaNotFoundError:
        error_msg = f"Schema not found for kind: {kind}."
        remediation_msg = "Use the `get_schema_mapping` tool to list available kinds."
        return await _log_and_return_error(ctx=ctx, error=error_msg, remediation=remediation_msg)
    except BranchNotFoundError as exc:
        return await _log_and_return_error(ctx=ctx, error=exc, remediation="Check the branch name or your permissions.")

    schema = await client.schema.get(kind=kind, branch=branch)

    return MCPResponse(
        status=MCPToolStatus.SUCCESS,
        data=schema.model_dump(),
    )


@mcp.tool(tags={"schemas", "retrieve"}, annotations=ToolAnnotations(readOnlyHint=True))
async def get_schemas(
    ctx: Context,
    branch: Annotated[
        str | None,
        Field(default=None, description="Branch to retrieve schemas from. Defaults to None (uses default branch)."),
    ],
    exclude_profiles: Annotated[
        bool, Field(default=True, description="Whether to exclude Profile schemas. Defaults to True.")
    ],
    exclude_templates: Annotated[
        bool, Field(default=True, description="Whether to exclude Template schemas. Defaults to True.")
    ],
) -> MCPResponse:
    """Retrieve all schemas from Infrahub, optionally excluding Profiles and Templates.

    Parameters:
        infrahub_client: Infrahub client to use
        branch: Branch to retrieve schemas from
        exclude_profiles: Whether to exclude Profile schemas. Defaults to True.
        exclude_templates: Whether to exclude Template schemas. Defaults to True.

    Returns:
        Dictionary with success status and schemas.

    """
    client: InfrahubClient = ctx.request_context.lifespan_context.client
    await ctx.info(f"Fetching all schemas in branch {branch or 'main'} from Infrahub...")

    try:
        all_schemas = await client.schema.all(branch=branch)
    except BranchNotFoundError as exc:
        return await _log_and_return_error(ctx=ctx, error=exc, remediation="Check the branch name or your permissions.")

    # Filter out Profile and Template if requested
    filtered_schemas = {}
    for kind, schema in all_schemas.items():
        if (exclude_templates and schema.namespace == "Template") or (
            exclude_profiles and schema.namespace == "Profile"
        ):
            continue
        filtered_schemas[kind] = schema.model_dump()

    return MCPResponse(
        status=MCPToolStatus.SUCCESS,
        data=filtered_schemas,
    )



from typing import TYPE_CHECKING, Annotated, Any

from fastmcp import Context, FastMCP
from infrahub_sdk.exceptions import GraphQLError, SchemaNotFoundError
from infrahub_sdk.types import Order
from mcp.types import ToolAnnotations
from pydantic import Field

from infrahub_mcp.constants import schema_attribute_type_mapping
from infrahub_mcp.utils import MCPResponse, MCPToolStatus, _log_and_return_error, convert_node_to_dict

if TYPE_CHECKING:
    from infrahub_sdk.client import InfrahubClient

mcp: FastMCP = FastMCP(name="Infrahub Nodes")


@mcp.tool(tags={"nodes", "retrieve"}, annotations=ToolAnnotations(readOnlyHint=True))
async def get_nodes(
    ctx: Context,
    kind: Annotated[str, Field(description="Kind of the objects to retrieve.")],
    branch: Annotated[
        str | None,
        Field(default=None, description="Branch to retrieve the objects from. Defaults to None (uses default branch)."),
    ],
    filters: Annotated[dict[str, Any] | None, Field(default=None, description="Dictionary of filters to apply.")],
    partial_match: Annotated[bool, Field(default=False, description="Whether to use partial matching for filters.")],
) -> MCPResponse:
    """Get all objects of a specific kind from Infrahub.

    To retrieve the list of available kinds, use the `get_schema_mapping` tool.
    To retrieve the list of available filters for a specific kind, use the `get_node_filters` tool.

    Parameters:
        kind: Kind of the objects to retrieve.
        branch: Branch to retrieve the objects from. Defaults to None (uses default branch).
        filters: Dictionary of filters to apply.
        partial_match: Whether to use partial matching for filters.

    Returns:
        MCPResponse with success status and objects.

    """
    client: InfrahubClient = ctx.request_context.lifespan_context.client
    await ctx.info(f"Fetching nodes of kind: {kind} with filters: {filters} from Infrahub...")

    # Verify if the kind exists in the schema and guide Tool if not
    try:
        schema = await client.schema.get(kind=kind, branch=branch)
    except SchemaNotFoundError:
        error_msg = f"Schema not found for kind: {kind}."
        remediation_msg = "Use the `get_schema_mapping` tool to list available kinds."
        return await _log_and_return_error(ctx=ctx, error=error_msg, remediation=remediation_msg)

    # TODO: Verify if the filters are valid for the kind and guide Tool if not

    try:
        if filters:
            await ctx.debug(f"Applying filters: {filters} with partial_match={partial_match}")
            nodes = await client.filters(
                kind=schema.kind,
                branch=branch,
                partial_match=partial_match,
                parallel=True,
                order=Order(disable=True),
                populate_store=True,
                prefetch_relationships=True,
                **filters,
            )
        else:
            nodes = await client.all(
                kind=schema.kind,
                branch=branch,
                parallel=True,
                order=Order(disable=True),
                populate_store=True,
                prefetch_relationships=True,
            )
    except GraphQLError as exc:
        return await _log_and_return_error(ctx=ctx, error=exc, remediation="Check the provided filters or the kind name.")

    # Format the response with serializable data
    # serialized_nodes = []
    # for node in nodes:
    #     node_data = await convert_node_to_dict(obj=node, branch=branch)
    #     serialized_nodes.append(node_data)
    serialized_nodes = [obj.display_label for obj in nodes]

    # Return the serialized response
    await ctx.debug(f"Retrieved {len(serialized_nodes)} nodes of kind {kind}")

    return MCPResponse(
        status=MCPToolStatus.SUCCESS,
        data=serialized_nodes,
    )


@mcp.tool(tags={"nodes", "filters", "retrieve"}, annotations=ToolAnnotations(readOnlyHint=True))
async def get_node_filters(
    ctx: Context,
    kind: Annotated[str, Field(description="Kind of the objects to retrieve.")],
    branch: Annotated[
        str | None,
        Field(default=None, description="Branch to retrieve the objects from. Defaults to None (uses default branch)."),
    ],
) -> MCPResponse:
    """Retrieve all the available filters for a specific schema node kind.

    There's multiple types of filters
    attribute filters are in the form attribute__value

    relationship filters are in the form relationship__attribute__value
    you can find more information on the peer node of the relationship using the `get_schema` tool

    Filters that start with parent refer to a related generic schema node.
    You can find the type of that related node by inspected the output of the `get_schema` tool.

    Parameters:
        kind: Kind of the objects to retrieve.
        branch: Branch to retrieve the objects from. Defaults to None (uses default branch).

    Returns:
        MCPResponse with success status and filters.
    """
    client: InfrahubClient = ctx.request_context.lifespan_context.client
    await ctx.info(f"Fetching available filters for kind: {kind} from Infrahub...")

    # Verify if the kind exists in the schema and guide Tool if not
    try:
        schema = await client.schema.get(kind=kind, branch=branch)
    except SchemaNotFoundError:
        error_msg = f"Schema not found for kind: {kind}."
        remediation_msg = "Use the `get_schema_mapping` tool to list available kinds."
        return await _log_and_return_error(ctx=ctx, error=error_msg, remediation=remediation_msg)

    filters = {
        f"{attribute.name}__value": schema_attribute_type_mapping.get(attribute.kind, "String")
        for attribute in schema.attributes
    }

    for relationship in schema.relationships:
        relationship_schema = await client.schema.get(kind=relationship.peer)
        relationship_filters = {
            f"{relationship.name}__{attribute.name}__value": schema_attribute_type_mapping.get(attribute.kind, "String")
            for attribute in relationship_schema.attributes
        }
        filters.update(relationship_filters)

    return MCPResponse(
        status=MCPToolStatus.SUCCESS,
        data=filters,
    )


@mcp.tool(tags={"nodes", "retrieve"}, annotations=ToolAnnotations(readOnlyHint=True))
async def get_related_nodes(
    ctx: Context,
    kind: Annotated[str, Field(description="Kind of the objects to retrieve.")],
    relation: Annotated[str, Field(description="Name of the relation to fetch.")],
    filters: Annotated[dict[str, Any] | None, Field(default=None, description="Dictionary of filters to apply.")],
    branch: Annotated[
        str | None,
        Field(default=None, description="Branch to retrieve the objects from. Defaults to None (uses default branch)."),
    ],
) -> MCPResponse:
    """Retrieve related nodes by relation name and a kind.

    Args:
        kind: Kind of the node to fetch.
        filters: Filters to apply on the node to fetch.
        relation: Name of the relation to fetch.
        branch: Branch to fetch the node from. Defaults to None (uses default branch).

    Returns:
        MCPResponse with success status and objects.

    """
    client: InfrahubClient = ctx.request_context.lifespan_context.client
    filters = filters or {}
    if branch:
        await ctx.info(f"Fetching nodes related to {kind} with filters {filters} in branch {branch} from Infrahub...")
    else:
        await ctx.info(f"Fetching nodes related to {kind} with filters {filters} from Infrahub...")

    try:
        node_id = node_hfid = None
        if filters.get("ids"):
            node_id = filters["ids"][0]
        elif filters.get("hfid"):
            node_hfid = filters["hfid"]
        if node_id:
            node = await client.get(
                kind=kind,
                id=node_id,
                branch=branch,
                include=[relation],
                prefetch_relationships=True,
                populate_store=True,
            )
        elif node_hfid:
            node = await client.get(
                kind=kind,
                hfid=node_hfid,
                branch=branch,
                include=[relation],
                prefetch_relationships=True,
                populate_store=True,
            )
    except Exception as exc:  # noqa: BLE001
        return await _log_and_return_error(ctx=ctx, error=exc)

    rel = getattr(node, relation, None)
    if not rel:
        return await _log_and_return_error(
            ctx=ctx,
            error=f"Relation '{relation}' not found in kind '{kind}'.",
            remediation="Check the schema for the kind to confirm if the relation exists.",
        )
    peers = [
        await convert_node_to_dict(
            branch=branch,
            obj=peer.peer,
            include_id=True,
        )
        for peer in rel.peers
    ]

    return MCPResponse(
        status=MCPToolStatus.SUCCESS,
        data=peers,
    )



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from fastmcp import FastMCP
from infrahub_sdk.client import InfrahubClient

from infrahub_mcp.tools.branch import mcp as branch_mcp
from infrahub_mcp.tools.gql import mcp as graphql_mcp
from infrahub_mcp.tools.nodes import mcp as nodes_mcp
from infrahub_mcp.tools.schema import mcp as schema_mcp


@dataclass
class AppContext:
    client: InfrahubClient


@asynccontextmanager
async def app_lifespan(server: FastMCP) -> AsyncIterator[AppContext]:  # noqa: ARG001, RUF029
    """Manages application lifecycle with type-safe context for the FastMCP server."""
    client = InfrahubClient()
    try:
        yield AppContext(client=client)
    finally:
        pass


mcp: FastMCP = FastMCP(name="Infrahub MCP Server", version="0.1.0", lifespan=app_lifespan)

# Mount the various MCPs to the main server
mcp.mount(branch_mcp)
mcp.mount(graphql_mcp)
mcp.mount(nodes_mcp)
mcp.mount(schema_mcp)



import pytest
from agents import Runner
from agents.mcp import MCPServerStdio
from deepeval import assert_test
from deepeval.dataset.golden import Golden
from deepeval.metrics.answer_relevancy.answer_relevancy import AnswerRelevancyMetric
from deepeval.test_case.llm_test_case import LLMTestCase, ToolCall

from tests.utils import agent_context, extract_tools

goldens = [
    # Golden(
    #     name="find_devices",
    #     input="What are the devices of role edge",
    #     expected_output="""
    #         The devices with the role "edge" present in InfraHub are:
    #             atl1-edge1
    #             atl1-edge2
    #             den1-edge1
    #             den1-edge2
    #             dfw1-edge1
    #             dfw1-edge2
    #             jfk1-edge1
    #             jfk1-edge2
    #             ord1-edge1
    #             ord1-edge2
    #         """,
    #     # expected_tools=[
    #     #     ToolCall(name="weather_forecast", input_parameters={"location": "Paris"})
    #     # ]
    # ),
    Golden(
        name="find_kind",
        input="what is the proper kind for a device",
        expected_output="The proper kind for a device is InfraDevice.",
        expected_tools=[ToolCall(name="get_schema_mapping", input_parameters=None)],
    ),
]


@pytest.mark.parametrize("golden", goldens)
async def test_llm_app(local_mcp_server: MCPServerStdio, main_prompt: str, golden: Golden) -> None:
    # NOTE: it should be possible to mode agent_context to a fixture, need to investigate
    async with agent_context(
        name="Assistant",
        instructions=main_prompt,
        mcp_servers=[local_mcp_server],
    ) as agent:
        result = await Runner.run(agent, input=golden.input)

        # Extract the `tools` called during the run
        # This is useful to check if the agent called the expected tools
        # and to compare the input parameters of the tools with the expected ones
        # Currently, the method doesn't return the output of the tools, need to investigate if it's possible
        tools_called = extract_tools(result)

        test_case = LLMTestCase(
            name=golden.name,
            input=golden.input,
            actual_output=result.final_output,
            tools_called=tools_called,
            expected_tools=golden.expected_tools,
        )

        assert_test(
            test_case=test_case,
            metrics=[
                AnswerRelevancyMetric(),
                # ToolCorrectnessMetric(evaluation_params=[ToolCallParams.INPUT_PARAMETERS], strict_mode=True, should_exact_match=True)
            ],
        )



import sys
from pathlib import Path

from invoke import Context, task

CURRENT_DIRECTORY = Path(__file__).resolve()
DOCUMENTATION_DIRECTORY = CURRENT_DIRECTORY.parent / "docs"

MAIN_DIRECTORY_PATH = Path(__file__).parent


@task(name="format")
def format_all(context: Context) -> None:
    """Run RUFF to format all Python files."""

    exec_cmds = ["uv run ruff format .", "uv run ruff check src/ --fix"]
    with context.cd(MAIN_DIRECTORY_PATH):
        for cmd in exec_cmds:
            context.run(cmd)


@task
def lint_yaml(context: Context) -> None:
    """Run Linter to check all Python files."""
    print(" - Check code with yamllint")
    exec_cmd = "uv run yamllint ."
    with context.cd(MAIN_DIRECTORY_PATH):
        context.run(exec_cmd)


@task
def lint_mypy(context: Context) -> None:
    """Run Linter to check all Python files."""
    print(" - Check code with mypy")
    exec_cmd = "uv run mypy --show-error-codes src/{PYTHON_MODULE}"
    with context.cd(MAIN_DIRECTORY_PATH):
        context.run(exec_cmd)


@task
def lint_pylint(context: Context) -> None:
    """Run pylint against Emma pages."""
    print(" - Check code with pylint")
    exec_cmd = "uv run pylint src/*.py"
    with context.cd(MAIN_DIRECTORY_PATH):
        context.run(exec_cmd)


@task
def lint_ruff(context: Context) -> None:
    """Run Linter to check all Python files."""
    print(" - Check code with ruff")
    exec_cmd = "uv run ruff check src/"
    with context.cd(MAIN_DIRECTORY_PATH):
        context.run(exec_cmd)


@task(name="lint")
def lint_all(context: Context) -> None:
    """Run all linters."""
    lint_yaml(context)
    lint_ruff(context)
    lint_mypy(context)
    lint_pylint(context)


@task(name="docs")
def docs_build(context: Context) -> None:
    """Build documentation website."""
    exec_cmd = "npm run build"

    with context.cd(DOCUMENTATION_DIRECTORY):
        output = context.run(exec_cmd)

    if output.exited != 0:
        sys.exit(-1)



from fastmcp import Client

from infrahub_mcp.server import mcp


async def test_list_schema() -> None:
    async with Client(mcp) as client:
        result = await client.call_tool("get_schema_mapping")
        assert isinstance(result.data, dict)
        assert "LocationSite" in result.data


async def test_get_node_filters(locationsite_filters: dict[str, str]) -> None:
    async with Client(mcp) as client:
        result = await client.call_tool("get_node_filters", {"kind": "LocationSite"})
        assert isinstance(result.data, dict)
        assert result.data == locationsite_filters


async def test_get_nodes() -> None:
    async with Client(mcp) as client:
        result = await client.call_tool("get_nodes", {"kind": "LocationSite"})
        assert isinstance(result.data, list)
        assert result.data == [
            "atl1",
            "den1",
            "dfw1",
            "jfk1",
            "ord1",
        ]



from enum import Enum
from pathlib import Path
from typing import Any, TypeVar

from fastmcp import Context
from infrahub_sdk.node import Attribute, InfrahubNode, RelatedNode, RelationshipManager
from pydantic import BaseModel

CURRENT_DIRECTORY = Path(__file__).parent.resolve()
PROMPTS_DIRECTORY = CURRENT_DIRECTORY / "prompts"


T = TypeVar("T")


class MCPToolStatus(Enum):
    SUCCESS = "success"
    ERROR = "error"


class MCPResponse[T](BaseModel):
    status: MCPToolStatus
    data: T | None = None
    error: str | None = None
    remediation: str | None = None


def get_prompt(name: str) -> str:
    prompt_file = PROMPTS_DIRECTORY / f"{name}.md"
    if not prompt_file.exists():
        raise FileNotFoundError(f"Prompt file '{prompt_file}' does not exist.")
    return (PROMPTS_DIRECTORY / f"{name}.md").read_text()


async def _log_and_return_error(ctx: Context, error: str | Exception, remediation: str | None = None) -> MCPResponse:
    """Log an error and return a standardized error response."""
    if isinstance(error, Exception):
        error = str(error)
    await ctx.error(message=error)
    return MCPResponse(
        status=MCPToolStatus.ERROR,
        error=error,
        remediation=remediation,
    )


async def convert_node_to_dict(*, obj: InfrahubNode, branch: str | None, include_id: bool = True) -> dict[str, Any]:  # noqa: C901
    data = {}

    if include_id:
        data["index"] = obj.id or None

    for attr_name in obj._schema.attribute_names:  # noqa: SLF001
        attr: Attribute = getattr(obj, attr_name)
        data[attr_name] = str(attr.value)

    for rel_name in obj._schema.relationship_names:  # noqa: SLF001
        rel = getattr(obj, rel_name)
        if rel and isinstance(rel, RelatedNode):
            if not rel.initialized:
                await rel.fetch()
            related_node = obj._client.store.get(  # noqa: SLF001
                branch=branch,
                key=rel.peer.id,
                raise_when_missing=False,
            )
            if related_node:
                data[rel_name] = (
                    related_node.get_human_friendly_id_as_string(include_kind=True)
                    if related_node.hfid
                    else related_node.id
                )
        elif rel and isinstance(rel, RelationshipManager):
            peers: list[dict[str, Any]] = []
            if not rel.initialized:
                await rel.fetch()
            for peer in rel.peers:
                # FIXME: We are using the store to avoid doing to many queries to Infrahub
                # but we could end up doing store+infrahub if the store is not populated
                related_node = obj._client.store.get(  # noqa: SLF001
                    key=peer.id,
                    raise_when_missing=False,
                    branch=branch,
                )
                if not related_node:
                    await peer.fetch()
                    related_node = peer.peer
                peers.append(
                    related_node.get_human_friendly_id_as_string(include_kind=True)
                    if related_node.hfid
                    else related_node.id,
                )
            data[rel_name] = peers
    return data



from pathlib import Path

import pytest
from agents.mcp import MCPServerStdio, MCPServerStdioParams

from infrahub_mcp.utils import get_prompt

CURRENT_DIRECTORY = Path(__file__).parent.resolve()
ROOT_DIRECTORY = CURRENT_DIRECTORY.parent.parent.resolve()


@pytest.fixture(scope="session")
def main_prompt() -> str:
    return get_prompt("main")


@pytest.fixture(scope="session")
def local_mcp_server() -> MCPServerStdio:
    """Fixture to provide a local MCP server for testing."""

    return MCPServerStdio(
        name="infrahub",
        params=MCPServerStdioParams(
            command="uv",
            cwd=str(ROOT_DIRECTORY.absolute()),
            args=[
                "--directory",
                str(ROOT_DIRECTORY.absolute()),
                "run",
                "fastmcp",
                "run",
                "--no-banner",
                "src/infrahub_mcp/server.py:mcp",
            ],
            env={
                "INFRAHUB_ADDRESS": "https://sandbox.infrahub.app",
                "INFRAHUB_USERNAME": "otto",
                "INFRAHUB_PASSWORD": "infrahub",
            },
        ),
        cache_tools_list=True,
    )
